/*
 * Copyright (C) 2007 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <rtems.h>
#include <rtems/error.h>

#include "dlfcn.h"
#include "linker.h"

#define SYSSYMFILE	"sym.map.gz"
/* This file hijacks the symbols stubbed out in libdl.so. */

#define DL_SUCCESS                    0
#define DL_ERR_CANNOT_FIND_LIBRARY    1
#define DL_ERR_INVALID_LIBRARY_HANDLE 2
#define DL_ERR_BAD_SYMBOL_NAME        3
#define DL_ERR_SYMBOL_NOT_FOUND       4
#define DL_ERR_SYMBOL_NOT_GLOBAL      5

static const char *dl_errors[] = {
    [DL_SUCCESS] = NULL,
    [DL_ERR_CANNOT_FIND_LIBRARY] = "Cannot find library",
    [DL_ERR_INVALID_LIBRARY_HANDLE] = "Invalid library handle",
    [DL_ERR_BAD_SYMBOL_NAME] = "Invalid symbol name",
    [DL_ERR_SYMBOL_NOT_FOUND] = "Symbol not found",
    [DL_ERR_SYMBOL_NOT_GLOBAL] = "Symbol is not global",
};

static int dl_last_err = DL_SUCCESS;

#define likely(expr)   __builtin_expect (expr, 1)
#define unlikely(expr) __builtin_expect (expr, 0)

#define RTEMS_MUTEX_ATTRIBS \
  (RTEMS_PRIORITY | RTEMS_SIMPLE_BINARY_SEMAPHORE | \
   RTEMS_NO_INHERIT_PRIORITY | RTEMS_NO_PRIORITY_CEILING | RTEMS_LOCAL)

static rtems_id dl_lock_id;

static void dl_lock(void)
{
  rtems_status_code sc = rtems_semaphore_obtain (dl_lock_id, RTEMS_WAIT, 0);
  if (sc != RTEMS_SUCCESSFUL)
    printf ("dlfnc: mutex: obtain failed: %s\n", 
            rtems_status_text (sc));
}

static void dl_unlock(void)
{
  rtems_status_code sc = rtems_semaphore_release (dl_lock_id);
  if (sc != RTEMS_SUCCESSFUL)
    printf ("dlfnc: mutex: release failed: %s\n", 
            rtems_status_text (sc));
}

void *dlopen(const char *filename, int flag) 
{
	soinfo *ret;
	static int initialized = 0;

	if (!initialized) {
    rtems_status_code sc;
    sc = rtems_semaphore_create(rtems_build_name ('D', 'L', 'F', 'N'),
                                1, RTEMS_MUTEX_ATTRIBS, 0, &dl_lock_id);
    if (sc != RTEMS_SUCCESSFUL) {
      printf("dlfnc: cannot create lock: %s", rtems_status_text (sc));
      return NULL;
    }
		__linker_init(SYSSYMFILE);
		initialized = 1;
	}
	
	dl_lock();

	ret = find_library(filename);
	if (unlikely(ret == NULL)) {
		dl_last_err = DL_ERR_CANNOT_FIND_LIBRARY;
	} else {
		ret->refcount++;
	}
	dl_unlock();
	return ret;
}

const char *dlerror(void)
{
    const char *err = dl_errors[dl_last_err];
    dl_last_err = DL_SUCCESS;
    return err;
}

void *dlsym(void *handle, const char *symbol)
{
    unsigned long sym;

    dl_lock();
    
    if(unlikely(handle == 0)) { 
        dl_last_err = DL_ERR_INVALID_LIBRARY_HANDLE;
        goto err;
    }
    if(unlikely(symbol == 0)) {
        dl_last_err = DL_ERR_BAD_SYMBOL_NAME;
        goto err;
    }
    
    if(handle == RTLD_DEFAULT) {
        sym = lookup(symbol);
    } else if(handle == RTLD_NEXT) {
        sym = lookup(symbol);
    } else {
        sym = lookup_in_library((soinfo*) handle, symbol);
    }
    
    if(likely(sym != 0)) {
        dl_unlock();
        return (void*)sym;
    }
    else dl_last_err = DL_ERR_SYMBOL_NOT_FOUND;

err:
    dl_unlock();
    return 0;
}

int dlclose(void *handle)
{
	dl_lock();
	(void)unload_library((soinfo*)handle);
	dl_unlock();
	return 0;
}
